#!/bin/bash
export GIT_PROTO=https
export GIT_SERVER=github.com
export GIT_PROJECT_NAME=empty
export GIT_PROJECT_DIR=second
export GIT_USER=""
export GIT_PASSWORD=""
export GIT_ORG_NAME="riotkit-org"
export WRITABLE_DIRS="files store cache"

# this example intentionally has no any hooks defined
